@@ -1,10 +1,10 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "mistty" "1.5"
+(define-package "mistty" "20250716.1914"
   "Shell/Comint alternative based on term.el."
   '((emacs "29.1"))
-  :url "https://github.com/szermatt/mistty"
-  :commit "e12c6af9894d6dbfd127047663275c763cab4e89"
-  :revdesc "e12c6af9894d"
+  :url "http://github.com/szermatt/mistty"
+  :commit "bfb17611cff6c845270050a0756a38489cdf4ed6"
+  :revdesc "bfb17611cff6"
   :keywords '("convenience" "unix")
   :authors '(("Stephane Zermatten" . "szermatt@gmx.net"))
   :maintainers '(("Stephane Zermatten" . "szermatt@gmx.net")))
