@@ -68,8 +68,7 @@ possible problem:
   ;; The process the queue is communicating with.
   proc
 
-  ;; A `mistty--interact' that yields strings to send to the terminal
-  ;; or nil.
+  ;; A `mistty--interact' that communicates with the terminal or nil.
   interact
 
   ;; A queue of `mistty--interact' to use after INTERACT.
@@ -94,15 +93,11 @@ possible problem:
   ;; Symbol that identifies the type of interaction.
   type
 
-  ;; Process to interact with. This is set when the interaction is
-  ;; enqueued.
-  proc
-
   ;; Callback function that will handle the next call to
   ;; mistty--interact-next. It takes a single argument.
   ;;
   ;; Such callbacks must never return a value. They must instead call
-  ;; mistty--interact-return, mistty--interact-wait-for-output or
+  ;; mistty--interact-return, mistty--interact-wait-for-output-then or
   ;; mistty--interact-done, which both throw 'mistty--interact-return.
   ;;
   ;; CB initially runs within the buffer that was current when
@@ -123,13 +118,12 @@ possible problem:
   ;; - 'stable once the process stops sending data for a certain time
   ;; - 'timeout if the process did not send anything for a long time
   ;;
-  ;; CB then raises 'mistty--interact-return with one of the following
-  ;; values:
-  ;; - a non-empty string to send to the terminal
-  ;; - 'wait-for-output to just wait for more output
-  ;; - 'done once the interaction is done. It'll be closed and discarded.
-  ;; (This is normally done by calling the function mistty--interact-return,
-  ;; mistty--interact-done or mistty--interact-wait-for-output.)
+  ;; CB then either:
+  ;; - throw 'mistty--interact-return 'done to end the interaction
+  ;; - throw 'mistty--interact-return <callable> to have <callable>
+  ;;   be called once new terminal data is available
+  ;; (This is normally done by calling the function
+  ;; mistty--interact-done or mistty--interact-wait-for-output-then.)
   cb
 
   ;; A function that releases any resource held during the
@@ -146,7 +140,17 @@ possible problem:
 
   ;; The buffer that was current when `mistty--make-interact' was
   ;; called. It is set before calling CLEANUP.
-  initial-buf)
+  initial-buf
+
+  ;; List of strings that should be sent to the process.
+  ;;
+  ;; As new strings are just pushed to this list, the strings are to
+  ;; be sent in reverse order.
+  ;;
+  ;; New strings are added by `mistty--interact-send' and consumed
+  ;; after CB returned and the decoder is in a state where calling
+  ;; `process-send-string' is ok. See #57 for details.
+  pending-output)
 
 (cl-defmacro mistty--interact (type (var) &rest body)
   "Create and return an interaction.
@@ -165,6 +169,17 @@ Returns the newly-created interaction."
            (lambda (&optional _) ,@body))
      ,var))
 
+(defun mistty--interact-take-cb (interact)
+  "Consume a callback from INTERACT.
+
+The callback is supposed to throw either a new callback, to replace this
+one, or \\='done, to end the interaction."
+  (prog1 (mistty--interact-cb interact)
+    (setf (mistty--interact-cb interact)
+          (let ((type (mistty--interact-type interact)))
+            (lambda (&optional _)
+              (error "Already consumed interact-cb of %s" type))))))
+
 (defsubst mistty--interact-callable-p (interact)
   "Return non-nil if INTERACT is callable."
   (functionp (mistty--interact-call interact)))
@@ -182,14 +197,14 @@ be pass to a :then argument in `mistty--interact-return'."
 
 (defun mistty--queue-interact-type (queue)
   "Return the type of the QUEUE current interact or nil."
-  (when-let ((interact (mistty--queue-interact queue)))
+  (when-let* ((interact (mistty--queue-interact queue)))
     (mistty--interact-type interact)))
 
 (defun mistty--queue-last-interact (queue)
   "Return the last interact of QUEUE or nil.
 
 The last interact often is the one that's currently running."
-  (if-let ((more (mistty--queue-more-interacts queue)))
+  (if-let* ((more (mistty--queue-more-interacts queue)))
       (car (last more))
     (mistty--queue-interact queue)))
 
@@ -197,7 +212,7 @@ The last interact often is the one that's currently running."
   "Send STR to PROC, if it is still live."
   (when (and (mistty--nonempty-str-p str)
              (process-live-p proc))
-    (mistty-log "SEND[%s]" str)
+    (mistty-log "SEND %S" str)
     (process-send-string proc str)))
 
 (defun mistty--enqueue-str (queue str &optional fire-and-forget)
@@ -211,8 +226,7 @@ Does nothing is STR is nil or empty."
        (mistty--interact-send interact str)
        (if fire-and-forget
            (mistty--interact-done)
-         (mistty--interact-return-then
-          interact #'mistty--interact-done))))))
+         (mistty--interact-wait-for-output-then #'mistty--interact-done))))))
 
 (defun mistty--enqueue (queue interact &optional prepend)
   "Add INTERACT to QUEUE.
@@ -233,8 +247,6 @@ right away.
 Does nothing if INTERACT is nil."
   (cl-assert (mistty--queue-p queue))
   (when interact
-    (setf (mistty--interact-proc interact)
-          (mistty--queue-proc queue))
     (cond
      ;; This is the first mistty-interact; kick things off.
      ((mistty--queue-empty-p queue)
@@ -256,47 +268,59 @@ Does nothing if INTERACT is nil."
 If VALUE is set, send that value to the first call to
 `mistty--interact-next'."
   (cl-assert (mistty--queue-p queue))
-  (mistty--dequeue-1 queue value)
-  (unless (mistty--queue-empty-p queue)
-    (setf (mistty--queue-timeout queue)
-          (run-with-timer
-           mistty-timeout-s nil #'mistty--timeout-handler
-           (current-buffer) queue))))
+  (let ((pending-outputs (mistty--dequeue-1 queue value)))
+    (unless (mistty--queue-empty-p queue)
+      (setf (mistty--queue-timeout queue)
+            (run-with-timer
+             mistty-timeout-s nil #'mistty--timeout-handler
+             (current-buffer) queue)))
+
+    ;; Send pending output to the process now. This is done at the end
+    ;; once the queue and its interacts have been fully updated, in
+    ;; case process-send-string calls the output filter - and this
+    ;; code - again recursively. See issue #57
+    (let ((proc (mistty--queue-proc queue)))
+      (dolist (p (nreverse pending-outputs))
+        (dolist (str (nreverse p))
+          (process-send-string proc str))))))
 
 (cl-defun mistty--dequeue-1 (queue value)
   "Internal helper for `mistty--dequeue'.
 
 This function does all the work of `mistty-dequeue'. See its
-description for the meaning of QUEUE and VALUE."
-  (let ((proc (mistty--queue-proc queue)))
-    (mistty--cancel-timeout queue)
-    (while (mistty--queue-interact queue)
-      (pcase (condition-case-unless-debug err
-                 (mistty--interact-next (mistty--queue-interact queue) value)
-               (error
-                (mistty-log "Interaction failed; giving up: %s" err)
-                (message "mistty: Interaction failed; giving up: %s" err)
-                'done))
+description for the meaning of QUEUE and VALUE.
 
+It returns pending outputs, in the form of a list of list of strings to
+send, in reverse order."
+  (mistty--cancel-timeout queue)
+  (let ((pending-outputs nil))
+    (while-let ((interact (mistty--queue-interact queue)))
+      (pcase-exhaustive
+          (unwind-protect
+              (condition-case-unless-debug err
+                  (prog1 (mistty--interact-next interact value)
+                    (when-let* ((p (mistty--interact-pending-output interact)))
+                      (push p pending-outputs)))
+                (error
+                 (mistty-log "Interaction failed; giving up: %s" err)
+                 (message "mistty: Interaction failed; giving up: %s" err)
+                 'done))
+            (setf (mistty--interact-pending-output interact) nil))
+        ;; Move on to the next interact
         ('done
          (setq value nil)
-         (mistty--interact-close (mistty--queue-interact queue))
-         (setf (mistty--queue-interact queue)
-               (pop (mistty--queue-more-interacts queue)))
-         (when (mistty--queue-interact queue)
-           (mistty-log "NEXT %s" (mistty--queue-interact-type queue))))
+         (mistty--interact-close interact)
+         (let ((next-interact (pop (mistty--queue-more-interacts queue))))
+           (setf (mistty--queue-interact queue) next-interact)
+           (when next-interact
+             (mistty-log "NEXT %s" (mistty--interact-type next-interact)))))
 
-        ;; Keep waiting
-        ('wait-for-output
-         (cl-return-from mistty--dequeue-1))
+        ;; Keep waiting; exits the loop
+        ((and (pred functionp) cb)
+         (setf (mistty--interact-cb interact) cb)
+         (cl-return-from mistty--dequeue-1 pending-outputs))))
 
-        ;; Normal sequences
-        ((and (pred mistty--nonempty-str-p) str)
-         (mistty--send-string proc str)
-         (cl-return-from mistty--dequeue-1))
-
-        (invalid (error "Yielded invalid value: '%s'"
-                        invalid))))))
+    pending-outputs))
 
 (defun mistty--dequeue-with-timer (queue &optional value)
   "Call `mistty--dequeue' on QUEUE with VALUE on a timer.
@@ -378,7 +402,7 @@ This passes VAL to `mistty-interact-cb'."
   (with-current-buffer (mistty--interact-buf interact)
     (prog1
         (catch 'mistty--interact-return
-          (let ((ret (funcall (mistty--interact-cb interact) val)))
+          (let ((ret (funcall (mistty--interact-take-cb interact) val)))
             (error "Unexpected return value from %s: %s"
                    (mistty--interact-type interact) ret)))
       (setf (mistty--interact-buf interact) (current-buffer)))))
@@ -391,82 +415,74 @@ After this call, `mistty--interact-next' fails and
   (setf (mistty--interact-cb interact)
         (lambda (&optional _)
           (error "Interaction was closed")))
-  (when-let ((func (mistty--interact-cleanup interact)))
+  (when-let* ((func (mistty--interact-cleanup interact)))
     (setf (mistty--interact-cleanup interact) nil)
     (with-current-buffer (mistty--interact-initial-buf interact)
       (funcall func))))
 
-(defun mistty--interact-wrap-accept (accept-f)
-  "Decorates a condition for accepting a change.
-
-This is a helper for building `mistty--interact' instances. It wraps
-around ACCEPT-F, which should be a function that returns non-nil once
-the state of the output is acceptable.
-
-The wrapped function takes into account timeouts, intermediate
-states and stable states; it knows when to give up and return
-even though ACCEPT-F still doesn't return non-nil.
-
-It returns non-nil once interaction should continue:
- - ==\'accept if the state is correct
- - ==\'give-up if the state is incorrect but too much time has
-   passed
- - nil if the caller should keep waiting."
-  (let ((try-count 0))
-    (when (and mistty--report-issue-function (funcall accept-f))
-      (funcall mistty--report-issue-function 'already-matching))
-    (lambda (res)
-      (cond
-       ((funcall accept-f) 'accept)
-
-       ((and (eq res 'stable) (< try-count mistty-max-try-count))
-        (cl-incf try-count)
-        ;; keep waiting
-        nil)
-
-       ((eq res 'timeout)
-        (when mistty--report-issue-function
-          (funcall mistty--report-issue-function 'hard-timeout))
-        'give-up)
-
-       ((eq res 'stable)
-        (when mistty--report-issue-function
-          (funcall mistty--report-issue-function 'soft-timeout))
-        'give-up)))))
-
 (defun mistty--interact-send (interact str)
-  "Send STR to the process associated with INTERACT."
-  (mistty--send-string (mistty--interact-proc interact) str))
+  "Send STR to the process associated with INTERACT.
 
-(cl-defun mistty--interact-return-then (interact then &key pred on-timeout)
+The output is only sent at the end of `mistty-dequeue'."
+  (push str (mistty--interact-pending-output interact)))
+
+(cl-defun mistty--interact-wait-for-output-then (then &key pred on-timeout)
   "Return from current function and wait for process.
 
-This function waits for output from the INTERACT's process.
+This function waits for output from the current interact's process.
 
 PRED must be a function that returns non-nil once output has been
 received and the condition is met. It is wrapped with
 `mistty--interact-wrap-accept' before it is used.
 
-If THEN is set, this is the function that's called once output has been
-received and, if PRED is set, the condition has been met.
+THEN is the function that's called once output has been received and, if
+PRED is set, the condition has been met.
 
 If ON-TIMEOUT is set, and the interact times out waiting for PRED to
-succeed, call ON-TIMEOUT instead of THEN. ON-TIMEOUT is ignored if PRED is
-unset.
+succeed, call ON-TIMEOUT instead of THEN.
+
 
 This call gives back control to the queue."
-  (cond
-   (pred
-    (let ((accept-f (mistty--interact-wrap-accept pred))
-          (then (or then (mistty--interact-cb interact))))
-      (setf (mistty--interact-cb interact)
-            (lambda (value)
-              (pcase (funcall accept-f value)
-                ('accept (funcall then))
-                ('give-up (funcall (or on-timeout then)))
-                (_ (throw 'mistty--interact-return 'wait-for-output)))))))
-   (then (setf (mistty--interact-cb interact) then)))
-  (throw 'mistty--interact-return 'wait-for-output))
+  (unless (functionp then)
+    (error "not callable: %S" then))
+  (if (null pred)
+      (throw 'mistty--interact-return
+             (lambda (res)
+               (if (eq res 'timeout)
+                   (progn
+                     (when mistty--report-issue-function
+                       (funcall mistty--report-issue-function 'hard-timeout))
+                     (funcall (or on-timeout then)))
+                 (funcall then))))
+
+    ;; Wait for PRED to become true or for a timeout.
+    (when (and mistty--report-issue-function (funcall pred))
+      (funcall mistty--report-issue-function 'already-matching))
+    (let ((try-count 0))
+      (cl-labels
+          ((accept (res)
+             (cond
+              ((funcall pred) (funcall then))
+
+              ((eq res 'intermediate)
+               (throw 'mistty--interact-return #'accept))
+
+              ((and (eq res 'stable) (< try-count mistty-max-try-count))
+               (cl-incf try-count)
+               (throw 'mistty--interact-return #'accept))
+
+              ((eq res 'stable)
+               (when mistty--report-issue-function
+                 (funcall mistty--report-issue-function 'soft-timeout))
+               (funcall (or on-timeout then)))
+
+              ((eq res 'timeout)
+               (when mistty--report-issue-function
+                 (funcall mistty--report-issue-function 'hard-timeout))
+               (funcall (or on-timeout then)))
+
+              (t (error "Unexpected interact cb arg %s" res)))))
+        (throw 'mistty--interact-return #'accept)))))
 
 (defun mistty--call-interact (interact type &rest args)
   "Call INTERACT and pass it ARGS.
