@@ -25,12 +25,16 @@
 (defvar term-height) ; defined in term.el
 (defvar term-home-marker) ; defined in term.el
 
+(require 'pcase)
 (require 'subr-x)
 (eval-when-compile
   (require 'cl-lib))
 
 (require 'mistty-util)
 (require 'mistty-log)
+(require 'mistty-accum)
+(eval-when-compile
+  (require 'mistty-accum-macros))
 
 ;;; Code:
 
@@ -138,7 +142,7 @@ definition, because it doesn't allow editing what's above."
 (defconst mistty-del "\C-h"
   "Sequence to send to the process when backspace is pressed.
 
-Both BS (\\C-h) and DEL (\\d) have been used to map to backspace, though
+Both BS (^H) and DEL (^?) have been used to map to backspace, though
 inconsistently across terminals. Mistty 1.3 and earlier sent DEL, but
 that was switched to BS when Fish 4 started interpreting that as the
 delete key instead of the backspace key.
@@ -146,12 +150,39 @@ delete key instead of the backspace key.
 If you change this key, remember to change the mapping in
 `mistty-term-key-map' as well")
 
+(defcustom mistty-term-mode-hook (list #'mistty-call-term-mode-hook)
+  "Hook run in in term-mode buffers created by MisTTY.
+
+This hook overrides `term-mode-hook' for term buffers started by MisTTY
+to allow configuring MisTTY's term buffers differently from normal term
+buffers.
+
+The default includes `mistty-call-term-mode-hook', which calls the
+original `term-mode-hook'.
+
+If you'd like to have completely different configuration for normal term-mode
+buffers and term-mode buffers started by Mistty, call:
+
+  (remove-hook \\='mistty-term-mode-hook \\='mistty-call-term-mode-hook)
+
+You might want to execute the above command as well if you have reasons
+to think that some term-mode customization are interfering with MisTTY's
+operations."
+  :group 'mistty
+  :type 'hook)
+
+(defvar mistty-shadowed-term-mode-hook nil
+  "Special variable under which hooks found it `term-mode-hook' are stored.
+
+This is allows running `term-mode-hook' or not, from
+`mistty-term-mode-hook'.")
+
 (defvar mistty-term-key-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "<tab>") "\t")
     (define-key map (kbd "<return>") "\C-m")
     (define-key map (kbd "<backspace>") mistty-del)
-    (define-key map (kbd "<esc>") "\e")
+    (define-key map (kbd "<escape>") "\e")
 
     ;; The following is a reversed copy of xterm-function-map from
     ;; term/xterm. Simulating xterm keys is generally convenient, as
@@ -487,11 +518,6 @@ text is written to the terminal.
 
 This variable is available in the work buffer.")
 
-(defvar-local mistty--undecoded-bytes nil
-  "Bytes leftover in the last call to `mistty--emulate-terminal'.
-
-They'll be processed once more data is passed to the next call.")
-
 (defvar-local mistty--original-cursor nil
   "The local value `cursor-type' had before it was hidden.
 
@@ -515,6 +541,14 @@ accessible from either buffer.
 Always access it through the places `mistty--prompt'
 `mistty--prompt-archive' and `mistty--prompt-counter'.")
 
+(defconst mistty--prompt-regexp
+  "[^[:alnum:][:cntrl:][:blank:]][[:blank:]]$"
+  "Regexp used to identify prompts.
+
+New, empty lines that might be prompts are evaluated against this
+regexp. This regexp should match something that looks like the
+end of a prompt with no commands.")
+
 (cl-defstruct (mistty--prompt-cell
                (:constructor mistty--make-prompt-cell
                              (&aux (counter 0))))
@@ -553,21 +587,25 @@ Always access it through the places `mistty--prompt'
   end
 
   ;; Text of the prompt, used when source=regexp.
-  text)
+  text
+
+  ;; Position of the start of the user input, if known.
+  ;; This is a (cons scrolline column).
+  user-input-start)
 
 (defun mistty--prompt ()
   "Get the value of the current `mistty--prompt' struct or nil."
-  (when-let ((cell mistty--prompt-cell))
+  (when-let* ((cell mistty--prompt-cell))
     (mistty--prompt-cell-current cell)))
 
 (defun mistty--prompt-archive ()
   "Get the list of archived `mistty--prompt' structs."
-  (when-let ((cell mistty--prompt-cell))
+  (when-let* ((cell mistty--prompt-cell))
     (mistty--prompt-cell-archive cell)))
 
 (defun mistty--prompt-counter ()
   "Get the number of prompt instances created in this buffer."
-  (when-let ((cell mistty--prompt-cell))
+  (when-let* ((cell mistty--prompt-cell))
     (mistty--prompt-cell-counter cell)))
 
 (gv-define-setter mistty--prompt (val)
@@ -575,7 +613,7 @@ Always access it through the places `mistty--prompt'
 
 The old value, if any, is pushed into `mistty--prompt-archive'."
   `(progn
-     (when-let ((old (mistty--prompt-cell-current mistty--prompt-cell)))
+     (when-let* ((old (mistty--prompt-cell-current mistty--prompt-cell)))
        (push old (mistty--prompt-cell-archive mistty--prompt-cell)))
      (setf (mistty--prompt-cell-current mistty--prompt-cell) ,val)))
 
@@ -584,29 +622,23 @@ The old value, if any, is pushed into `mistty--prompt-archive'."
   `(setf (mistty--prompt-cell-archive mistty--prompt-cell) ,val))
 
 (defun mistty--prompt-contains (prompt scrolline)
-  "Return non-nil if PROMPT contains SCROLLINE."
+  "Return non-nil if SCROLLINE is inside of PROMPT."
   (and (>= scrolline (mistty--prompt-start prompt))
        (or (null (mistty--prompt-end prompt))
            (< scrolline (mistty--prompt-end prompt)))))
 
-(defun mistty--emulate-terminal (proc str work-buffer)
+(defun mistty--emulate-terminal (proc str)
   "Handle process output as a terminal would.
 
-This function accepts output from PROC included into STR and
-forwards them to `term-emulate-terminal'.
+This function accepts output from PROC included into STR and forwards
+them to `term-emulate-terminal' with some modified functions, fix some
+issues.
 
-Some special sequences are interrupted and pre-processed:
+It also logs everything it receives to mistty-log.
 
-- enabling and disabling bracketed paste, which means that
-bracketed paste strings can be sent to PROC. The state is stored
-into `mistty-bracketed-paste' in the buffer WORK-BUFFER.
-
-- making cursor visible or invisible in WORK-BUFFER
-
-- OSC sequences, either fed to `mistty-osc-handlers' or ignored."
-  (cl-letf ((inhibit-modification-hooks nil) ;; run mistty--after-change-on-term
-            (inhibit-read-only t) ;; allow modifications in char mode
-            (start 0)
+This is meant as a drop-in replacement for `term-emulate-terminal' in
+all situations, even when no work buffer is available."
+  (cl-letf ((inhibit-read-only t) ;; allow modifications in char mode
             ;; Using term-buffer-vertical-motion causes strange
             ;; issues; avoid it. Additionally, it's not actually
             ;; necessary since term.el adds newlines instead of
@@ -627,182 +659,213 @@ into `mistty-bracketed-paste' in the buffer WORK-BUFFER.
                             (not (eq ?\n (char-before (point-max)))))
                    (cl-decf res))
                  res)))
-            ((symbol-function 'term-delete-chars)
-             (lambda (count)
-               (let ((save-point (point)))
-                 (move-to-column (+ (term-current-column) count) t)
-                 (delete-region save-point (point)))))
             ((symbol-function 'term--handle-colors-list)
              (let ((real-handle-colors-list (symbol-function 'term--handle-colors-list)))
                (lambda (parameters)
                  (funcall real-handle-colors-list parameters)
                  (setq term-current-face
-                       (mistty--clear-term-face-value term-current-face)))))
-            ((symbol-function 'move-to-column)
-             (let ((orig (symbol-function 'move-to-column)))
-               (lambda (&rest args)
-                 (apply #'mistty--around-move-to-column orig args)))))
+                       (mistty--clear-term-face-value term-current-face))))))
+    (mistty-log "RECV %S" str)
+    (term-emulate-terminal proc str)
+
+    ;; MisTTY always wants the point at process mark, no matter what.
+    ;; term-mode is not so categorical and might sometimes lose sync
+    ;; during resizes.
     (mistty--with-live-buffer (process-buffer proc)
-      (when mistty--undecoded-bytes
-        (setq str (concat mistty--undecoded-bytes str))
-        (setq mistty--undecoded-bytes nil))
+        (goto-char (process-mark proc)))))
 
-      ;; Skip Application Keypad (DECPAM) / Normal Keypad (DECPNM)
-      ;; Issued by Fish 4+ but unsupported by term.el.
-      (setq str (replace-regexp-in-string "\e[=>]" "" str))
+(defun mistty--add-skip-unsupported (accum)
+  "Skip some unsupported terminal sequences that confuse term.el.
 
-      ;; Note on OSC content: ECMA 48 8.3.89 only allows 0x08-0x0d
-      ;; 0x20-7e. That would disallow all non-US-ASCII characters,
-      ;; often used in file names, which would then need to be
-      ;; encoded. This would be inconvenient and error-prone, so we
-      ;; disallow the US-ASCII characters disallowed by ECMA 48 and
-      ;; allow all non-US-ASCII chars (usually multibyte UTF-8).
-      (while (string-match "\e\\(?1:\\[\\?\\(?:2004\\|25\\)[hl]\\|\\]\\(?2:\\(?:\\(?3:[a-zA-Z0-9]+\\);\\)?[^\x00-\x07\x0e-\x1f\x7f]*\\)\\(?4:\e\\\\\\|\a\\|\\'\\)\\)\\|\\(?5: +\\)\r" str start)
-        (let ((ext (match-string 1 str))
-              (osc (match-string 2 str))
-              (osc-code (match-string 3 str))
-              (osc-terminator (match-string 4 str))
-              (prompt-sp-end (match-end 5))
-              (seq-start (match-beginning 0))
-              (seq-end (match-end 0)))
-          (cond
-           ((equal ext "[?2004h") ; enable bracketed paste
-            (term-emulate-terminal proc (substring str start seq-end))
-            (unless mistty-bracketed-paste
-              (let* ((prompt (mistty--prompt))
-                     (inhibit-modification-hooks t)
-                     (scrolline (mistty--term-scrolline)))
-                (when (or (null prompt)
-                          (not (mistty--prompt-contains prompt scrolline)))
-                  ;; zsh enables bracketed paste only after having printed
-                  ;; the prompt. Try to find the beginning of the prompt
-                  ;; from prompt_sp or assume a single-line prompt.
-                  (when-let ((real-start
-                              (catch 'mistty-prompt-start
-                                (dolist (i '(0 -1 -2 -3))
-                                  (let ((pos (pos-eol i)))
-                                    (when (and (zerop (mistty--prompt-counter))
-                                               (= pos 1) (> (point) (pos-bol)))
-                                      (mistty-log "extend first prompt [1-%s]" (point))
-                                      (throw 'mistty-prompt-start (point-min)))
-                                    (when (get-text-property pos 'mistty-prompt-sp)
-                                      (mistty-log "prompt_sp %s [%s-%s]" i (1+ pos) (point))
-                                      (remove-text-properties
-                                       pos (1+ pos) '(term-line-wrap t 'mistty-prompt-sp nil))
-                                      (throw 'mistty-prompt-start (1+ pos)))))))
-                             (eol (pos-eol)))
-                    (when (> eol real-start)
-                      ;; mistty--changed is only called when bracketed
-                      ;; paste is on; mark past sections of the prompt
-                      ;; as changed, including to the eol to cover
-                      ;; right prompts, also written before.
-                      (mistty--changed real-start eol))
-                    (setq scrolline (mistty--term-scrolline-at real-start)))
-                  (setq prompt (mistty--make-prompt 'bracketed-paste scrolline))
-                  (mistty-log "Detected %s prompt #%s [%s-]"
-                              (mistty--prompt-source prompt)
-                              (mistty--prompt-input-id prompt)
-                              (mistty--prompt-start prompt))
-                  (setf (mistty--prompt) prompt))
-                (unless (eq 'osc133 (mistty--prompt-source prompt))
-                  (setf (mistty--prompt-source prompt) 'bracketed-paste)
-                  (setf (mistty--prompt-end prompt) nil)))
-              (setq mistty-bracketed-paste t)
-              (mistty--with-live-buffer work-buffer
-                (setq mistty-bracketed-paste t))))
-          ((equal ext "[?2004l") ; disable bracketed paste
-           (term-emulate-terminal proc (substring str start seq-end))
-           (when mistty-bracketed-paste
-             (when-let ((prompt (mistty--prompt))
-                        (scrolline (if (eq ?\n (char-before (point)))
-                                       (mistty--term-scrolline)
-                                     (1+ (mistty--term-scrolline)))))
-               (when (and (eq 'bracketed-paste (mistty--prompt-source prompt))
-                          (null (mistty--prompt-end prompt))
-                          (> scrolline (mistty--prompt-start prompt)))
-                 (setf (mistty--prompt-end prompt) scrolline)))
-             (setq mistty-bracketed-paste nil)
-             (mistty--with-live-buffer work-buffer
-               (setq mistty-bracketed-paste nil))))
-          ((equal ext "[?25h") ; make cursor visible
-           (term-emulate-terminal proc (substring str start seq-end))
-           (mistty--with-live-buffer work-buffer
-             (mistty--show-cursor)))
-          ((equal ext "[?25l") ; make cursor invisible
-           (term-emulate-terminal proc (substring str start seq-end))
-           (mistty--with-live-buffer work-buffer
-             (mistty--hide-cursor)))
-          ((and osc (length= osc-terminator 0))
-           (term-emulate-terminal proc (substring str start seq-start))
-           ;; sequence is not finished; save it for later
-           (setq mistty--undecoded-bytes (substring str seq-start)))
-          (osc
-           (term-emulate-terminal proc (substring str start seq-start))
-           (when-let ((code osc-code)
-                      (handler (cdr (assoc-string code mistty-osc-handlers))))
-             (funcall handler code
-                      (decode-coding-string
-                       (string-remove-prefix (concat code ";") osc)
-                       locale-coding-system t))))
-          (prompt-sp-end
-           (term-emulate-terminal proc (substring str start prompt-sp-end))
-           (when (or (and (= (1- term-width) (term-current-column))
-                          (eq ?\  (char-before (point))))
-                     (and (get-text-property (pos-eol 0) 'term-line-wrap)
-                          (string-match "^ *$" (buffer-substring (pos-bol) (pos-eol)))))
-             (let ((inhibit-modification-hooks t)
-                   (pos (pos-eol 0)))
-               (put-text-property pos (1+ pos) 'mistty-prompt-sp t)))
-           ;; Include the \r in the next loop.
-           (setq seq-end prompt-sp-end)))
-        (setq start seq-end)))
-    (let ((split (mistty--split-incomplete-chars (substring str start))))
-      (when (length> (cdr split) 0)
-        (setq mistty--undecoded-bytes (cdr split)))
-      (term-emulate-terminal proc (car split)))))
+This function adds processors to ACCUM to skip Application
+Keypad (DECPAM) / Normal Keypad (DECPNM) Issued by Fish 4+ but just
+ecoed by term.el."
+  (mistty--accum-add-processor
+   accum
+   '(seq ESC (char "=>")) #'ignore))
 
-  (when-let ((change-start
-              (when (and mistty--term-changed
-                         (>= mistty--term-changed (point-min))
-                         (< mistty--term-changed (point-max)))
-                (text-property-any
-                 mistty--term-changed (point-max) 'mistty-changed t))))
-    (mistty--term-postprocess change-start term-width))
-  (setq mistty--term-changed nil))
+(defun mistty--add-osc-detection (accum)
+  "Handle OSC code in ACCUM.
 
-(defun mistty--split-incomplete-chars (str)
-  "Extract incomplete multibyte chars at the end of STR.
+Known OSC codes are passed down to handlers registered in
+`mistty-osc-handlers'."
+  (mistty--accum-add-processor-lambda accum
+      (ctx '(seq OSC (let code Ps) ?\; (let text Pt) ST))
+   (when-let* ((handler (cdr (assoc-string code mistty-osc-handlers))))
+     (mistty--accum-ctx-flush ctx)
+     (let ((inhibit-modification-hooks t)
+           (inhibit-read-only t))
+       (funcall handler code
+                (decode-coding-string text locale-coding-system t))))))
 
-This function detects multibyte chars that couldn't be decoded at
-the end of STR and splits it into a cons of complete string and
-remaining bytes.
+(defun mistty--add-prompt-detection (accum)
+  "Register processors to ACCUM for prompt detection.
 
-term.el is meant to do that, but it fails, because `char-charset'
-alone doesn't behave the way term.el assumes (anymore?). This is
-hopefully a temporary workaround."
-  (let* ((len (length str))
-         (end (substring str (max 0 (- len 8))))
-         (decoded-end (decode-coding-string end locale-coding-system t))
-         (undecoded-count 0)
-         (i (1- (length decoded-end))))
-    (while (and (>= i 0) (mistty--eight-bit-char-p decoded-end i))
-      (cl-incf undecoded-count)
-      (cl-decf i))
-    (if (zerop undecoded-count)
-        (cons str "")
-      (cons
-       (substring str 0 (- len undecoded-count))
-       (substring str (- len undecoded-count))))))
+Detected prompts can be found in `mistty-prompt'."
+  (mistty--accum-add-post-processor
+   accum #'mistty--term-postprocess-changed)
+  (mistty--accum-add-post-processor
+   accum (mistty--regexp-prompt-detector))
 
-(defun mistty--eight-bit-char-p (str index)
-  "Check whether char in STR at INDEX has been decoded."
-  ;; logic taken from Emacs 29 describe-char
-  (let ((c (aref str index)))
-    (eq 'eight-bit
-        (if (and (not enable-multibyte-characters) (>= c 128))
-            'eight-bit
-          (or (get-text-property index 'charset str)
-              (char-charset c))))))
+  ;; Enable bracketed paste
+  (mistty--accum-add-processor
+   accum
+   '(seq CSI "?2004h")
+   (lambda (ctx _)
+     (mistty--accum-ctx-flush ctx)
+     (unless mistty-bracketed-paste
+       (let* ((prompt (mistty--prompt))
+              (inhibit-read-only t)
+              (inhibit-modification-hooks t)
+              (scrolline (mistty--term-scrolline)))
+         (when (or (null prompt)
+                   (not (mistty--prompt-contains prompt scrolline)))
+           ;; zsh enables bracketed paste only after having printed
+           ;; the prompt. Try to find the beginning of the prompt
+           ;; from prompt_sp or assume a single-line prompt.
+           (when-let* ((real-start
+                       (catch 'mistty-prompt-start
+                         (dolist (i '(0 -1 -2 -3))
+                           (let ((pos (pos-eol i)))
+                             (when (and (zerop (mistty--prompt-counter))
+                                        (= pos 1) (> (point) (pos-bol)))
+                               (mistty-log "extend first prompt [1-%s]" (point))
+                               (throw 'mistty-prompt-start (point-min)))
+                             (when (get-text-property pos 'mistty-prompt-sp)
+                               (mistty-log "prompt_sp %s [%s-%s]" i (1+ pos) (point))
+                               (remove-text-properties
+                                pos (1+ pos) '(term-line-wrap t 'mistty-prompt-sp nil))
+                               (throw 'mistty-prompt-start (1+ pos)))))))
+                      (eol (pos-eol)))
+             (when (> eol real-start)
+               ;; mistty--changed is only called when bracketed
+               ;; paste is on; mark past sections of the prompt
+               ;; as changed, including to the eol to cover
+               ;; right prompts, also written before.
+               (mistty--changed real-start eol))
+             (setq scrolline (mistty--term-scrolline-at real-start)))
+           (setq prompt (mistty--make-prompt 'bracketed-paste scrolline))
+           (mistty--term-remove-prompt_sp prompt)
+           (mistty-log "Detected %s prompt #%s [%s-]"
+                       (mistty--prompt-source prompt)
+                       (mistty--prompt-input-id prompt)
+                       (mistty--prompt-start prompt))
+           (setf (mistty--prompt) prompt))
+         (unless (eq 'osc133 (mistty--prompt-source prompt))
+           (setf (mistty--prompt-source prompt) 'bracketed-paste)
+           (setf (mistty--prompt-end prompt) nil)))
+       (setq mistty-bracketed-paste t))))
+
+  ;; Disable bracketed paste
+  (mistty--accum-add-processor
+   accum
+   '(seq CSI "?2004l")
+   (lambda (ctx _)
+     (mistty--accum-ctx-flush ctx)
+     (when mistty-bracketed-paste
+       (when-let* ((prompt (mistty--prompt))
+                  (scrolline (if (eq ?\n (char-before (point)))
+                                 (mistty--term-scrolline)
+                               (1+ (mistty--term-scrolline)))))
+         (when (and (eq 'bracketed-paste (mistty--prompt-source prompt))
+                    (null (mistty--prompt-end prompt))
+                    (> scrolline (mistty--prompt-start prompt)))
+           (setf (mistty--prompt-end prompt) scrolline)))
+       (setq mistty-bracketed-paste nil))))
+
+  ;; Detect prompt-sp as many spaces followed by CR at the end of a
+  ;; line.
+  ;;
+  ;; Not using " \r" as regexp for the processor as it would mean
+  ;; waiting after a space in case a \r eventually comes. This isn't
+  ;; an escape sequence.
+  (mistty--accum-add-processor
+   accum
+   'CR
+   (lambda (ctx _)
+     ;; If we received at least 8 spaces before the \r (enough to fill
+     ;; the look-back buffer) flush and look at the state of the
+     ;; buffer just before the \r is taken into account.
+     (when (string= "        " (mistty--accum-ctx-look-back ctx))
+       (mistty--accum-ctx-flush ctx)
+       (when (or (and (= (1- term-width) (term-current-column))
+                      (eq ?\  (char-before (point))))
+                 (and (get-text-property (pos-eol 0) 'term-line-wrap)
+                      (string-match "^ *$" (buffer-substring (pos-bol) (pos-eol)))))
+         (let ((inhibit-modification-hooks t)
+               (inhibit-read-only t)
+               (pos (pos-eol 0)))
+           (put-text-property pos (1+ pos) 'mistty-prompt-sp t))))
+
+     (mistty--accum-ctx-push-down ctx "\r")))
+
+  ;; Detect and mark moves with mistty-maybe-skip
+  (mistty--accum-add-around-process-filter
+   accum
+   (lambda (func)
+     (cl-letf ((inhibit-modification-hooks nil) ;; run mistty--after-change-on-term
+               ((symbol-function 'term-delete-chars)
+                (lambda (count)
+                  (let ((save-point (point)))
+                    (move-to-column (+ (term-current-column) count) t)
+                    (delete-region save-point (point)))))
+               ((symbol-function 'move-to-column)
+                (let ((orig (symbol-function 'move-to-column)))
+                  (lambda (&rest args)
+                    (apply #'mistty--around-move-to-column orig args)))))
+       (funcall func)))))
+
+(defun mistty--term-remove-prompt_sp (prompt)
+  "Clear the mistty-prompt-sp property in PROMPT.
+
+This should be called for all newly-detected prompt to avoid confusing
+future prompts."
+  (when-let* ((start (mistty--term-scrolline-pos
+                     (mistty--prompt-start prompt))))
+    (when (< start (point-max))
+      (remove-text-properties (max (point-min) (1- start))
+                              (point-max) '
+                              (mistty-prompt-sp nil)))))
+
+(defun mistty--regexp-prompt-detector ()
+  "Build a post-processor that look for a new prompt at cursor.
+
+The return value is meant to be
+`mistty--accum-add-post-processor'.
+
+ The post-processor updates `mistty--prompt' after the content of the
+terminal buffer has been updated."
+  (let ((last-nonempty-scrolline 0))
+    (lambda ()
+      (let ((scrolline (mistty--term-scrolline)))
+        ;; Only look at new lines
+        (when (> scrolline
+                 (prog1 last-nonempty-scrolline
+                   ;; for next time
+                   (setq last-nonempty-scrolline
+                         (mistty--term-scrolline-at
+                          (mistty--last-non-ws)))))
+          (let ((cursor (point))
+                (bos (mistty--beginning-of-scrolline-pos))
+                (prompt (mistty--prompt)))
+            (when (and (or (null prompt)
+                           (and (mistty--prompt-end prompt)
+                                (>= scrolline (mistty--prompt-end prompt))))
+                       (> cursor bos)
+                       (>= cursor (mistty--last-non-ws))
+                       (string-match
+                        mistty--prompt-regexp
+                        (mistty--safe-bufstring bos cursor)))
+              (let ((prompt (mistty--make-prompt
+                             'regexp scrolline (1+ scrolline)
+                             :text (mistty--safe-bufstring bos (+ bos (match-end 0))))))
+                (setf (mistty--prompt) prompt)
+                (mistty-log "Suspected %s prompt #%s: [%s-%s] '%s'"
+                            (mistty--prompt-source prompt)
+                            (mistty--prompt-input-id prompt)
+                            (mistty--prompt-start prompt)
+                            (mistty--prompt-end prompt)
+                            (mistty--prompt-text prompt))))))))))
 
 (defun mistty-register-text-properties (id props)
   "Add PROPS to any text written to the terminal.
@@ -813,14 +876,14 @@ that off.
 If this function is called more than once with the same ID, only
 the last set of properties to be registered is applied."
   (unless (eq 'term-mode major-mode) (error "Requires a term-mode buffer"))
-  (if-let ((cell (assq id mistty--term-properties-to-add-alist)))
+  (if-let* ((cell (assq id mistty--term-properties-to-add-alist)))
       (setcdr cell props)
     (push (cons id props) mistty--term-properties-to-add-alist)))
 
 (defun mistty-unregister-text-properties (id)
   "Stop applying properties previously registered with ID."
   (unless (eq 'term-mode major-mode) (error "Requires a term-mode buffer"))
-  (when-let ((cell (assq id mistty--term-properties-to-add-alist)))
+  (when-let* ((cell (assq id mistty--term-properties-to-add-alist)))
     (setq mistty--term-properties-to-add-alist
           (delq cell
                 mistty--term-properties-to-add-alist))))
@@ -838,7 +901,9 @@ reported to the remote process.
 This function returns the newly-created buffer."
   (let ((term-buffer (generate-new-buffer name 'inhibit-buffer-hooks)))
     (with-current-buffer term-buffer
-      (term-mode)
+      (let* ((mistty-shadowed-term-mode-hook term-mode-hook)
+             (term-mode-hook mistty-term-mode-hook))
+        (term-mode))
       (font-lock-mode -1)
       (jit-lock-mode nil)
       (setq-local term-char-mode-buffer-read-only t)
@@ -866,7 +931,9 @@ This function returns the newly-created buffer."
         ;; prevents normal terminal resizing from working. This turns
         ;; it on again.
         (process-put proc 'adjust-window-size-function nil)
-        (set-process-window-size proc height width))
+        (set-process-window-size proc height width)
+        (set-process-filter proc (mistty--make-accumulator
+                                  #'mistty--emulate-terminal)))
       (setq-local term-raw-map local-map)
       (term-char-mode)
       (add-hook 'after-change-functions #'mistty--after-change-on-term nil t))
@@ -918,7 +985,7 @@ Must be called from the term buffer."
               (setq called t)
               ;; Set erase to ^H or ^? to stty so the terminal is
               ;; expecting the right delete value. Issue #12
-              (when-let ((stty-command (nth 1 program-args))
+              (when-let* ((stty-command (nth 1 program-args))
                          (erase-char (pcase mistty-del
                                        ("\C-h" "^H")
                                        ("\d" "^?"))))
@@ -956,7 +1023,7 @@ Must be called from the term buffer."
 BEG and END define the region that was modified."
   (let ((inhibit-modification-hooks t))
     (when (and mistty--term-properties-to-add-alist (> end beg))
-      (when-let ((props (apply #'append
+      (when-let* ((props (apply #'append
                                (mapcar #'cdr mistty--term-properties-to-add-alist))))
         ;; Merge sections with same properties separated by
         ;; whitespaces. The problem with setting text properties based
@@ -976,7 +1043,7 @@ BEG and END define the region that was modified."
       (mistty--changed beg end))))
 
 (defun mistty--changed (beg end)
-  "Mark text between BEG and AND as changed, forcing postprocess."
+  "Mark text between BEG and END as changed, forcing postprocess."
   (setq mistty--term-changed (if mistty--term-changed
                                  (min mistty--term-changed beg)
                                beg))
@@ -996,8 +1063,24 @@ arguments."
       (put-text-property
        initial-end (point) 'mistty-maybe-skip t))))
 
+(defun mistty--term-postprocess-changed ()
+  "Process mistty-maybe-skip text properties.
+
+This function turns mistty-maybe-skip into mistty-skip properties on the
+lines that have changed, as detected by `mistty--term-changed'."
+  (when (and mistty--term-changed (< mistty--term-changed (point-min)))
+    (setq mistty--term-changed (point-min)))
+  (when (and mistty--term-changed (>= mistty--term-changed (point-max)))
+    (setq mistty--term-changed nil))
+  (when-let* ((change-start
+              (when mistty--term-changed
+                (text-property-any
+                 mistty--term-changed (point-max) 'mistty-changed t))))
+    (mistty--term-postprocess change-start term-width))
+  (setq mistty--term-changed nil))
+
 (defun mistty--term-postprocess (region-start window-width)
-  "Sets mistty-skip and yank handlers after REGION-START.
+  "Set mistty-skip and yank handlers after REGION-START.
 
 WINDOW-WIDTH is used to detect right prompts.
 
@@ -1029,11 +1112,12 @@ detecting regions looking at a complete line."
 (defun mistty--detect-right-prompt (bol eol window-width)
   "Detect right prompt and return its left position or nil.
 
-BOL and EOL define the region to look in."
+BOL and EOL define the region to look in. WINDOW-WIDTH must be the width
+of the terminal, usually `term-width'."
   (let ((pos (1- eol)))
     (when (and (>= 3 (- window-width (mistty--line-width)))
                (not (get-text-property pos 'mistty-maybe-skip)))
-      (when-let ((first-maybe-skip (previous-single-property-change eol 'mistty-maybe-skip nil bol)))
+      (when-let* ((first-maybe-skip (previous-single-property-change eol 'mistty-maybe-skip nil bol)))
         (when (and (eq (char-before first-maybe-skip) ?\ )
                    (> first-maybe-skip bol))
           (setq pos (1- first-maybe-skip))
@@ -1043,13 +1127,8 @@ BOL and EOL define the region to look in."
             (cl-decf pos))
           (cl-incf pos)
           (add-text-properties
-           pos eol '(mistty-skip
-                     right-prompt
-                     yank-handler
-                     (nil "" nil nil)
-                     field right-prompt
-                     inhibit-line-move-field-capture t
-                     rear-nonsticky t))
+           pos eol '(mistty-skip right-prompt
+                     yank-handler (nil "" nil nil)))
 
           pos)))))
 
@@ -1182,7 +1261,7 @@ with every prompt if the env variable INSIDE_EMACS is set."
   (if (= (aref string 0) ?/)
       (let ((path (substring string 1)))
         (unless (file-remote-p path)
-          (when-let ((prefix (file-remote-p default-directory)))
+          (when-let* ((prefix (file-remote-p default-directory)))
             (setq path (concat prefix path))))
         ;; Not using cd here, to avoid a remote connection being made to
         ;; check the path.
@@ -1318,7 +1397,20 @@ Return nil if the row isn't reachable on the terminal."
   mistty--scrolline-base)
 
 (defun mistty-osc133 (_ osc-seq)
-  (when (length> osc-seq 0)
+  "Handle OSC 133 codes.
+
+OSC-SEQ contains the subcode followed optionally by a semi-colon and
+arguments (ignored).
+
+MisTTY supports code A-D:
+
+ - A marks the start of a new command.
+ - B marks the end of the prompt and the start of user input.
+ - C marks the start of command output.
+ - D marks the end of the command.
+
+Everything else is ignored."
+  (when (and (length> osc-seq 0) (not (term-using-alternate-sub-buffer)))
     (let ((command-char (aref osc-seq 0)))
       (pcase command-char
         (?A ;; start a new command
@@ -1326,21 +1418,20 @@ Return nil if the row isn't reachable on the terminal."
          ;; Overwrite any other prompt source.
          (let ((prompt (mistty--make-prompt 'osc133 (mistty--term-scrolline))))
            (setf (mistty--prompt) prompt)
+           (mistty--term-remove-prompt_sp prompt)
            (mistty-log "Detected %s prompt #%s [%s-]"
                        (mistty--prompt-source prompt)
                        (mistty--prompt-input-id prompt)
                        (mistty--prompt-start prompt))))
 
         (?B ;; end of prompt/start of user input
-         (when (> (point) (pos-bol))
-           (add-text-properties (pos-bol) (point)
-                                '(field prompt
-                                  inhibit-line-move-field-capture t
-                                  front-sticky (field inhibit-line-move-field-capture)
-                                  rear-nonsticky t))))
+         (when-let* ((prompt (mistty--prompt)))
+           (setf (mistty--prompt-user-input-start prompt)
+                 (cons (mistty--term-scrolline)
+                       (term-current-column)))))
 
         (?C ;; start of command output
-         (when-let ((prompt (mistty--prompt)))
+         (when-let* ((prompt (mistty--prompt)))
            (when (eq 'osc133 (mistty--prompt-source prompt))
              (mistty-log "Closed %s prompt #%s [%s-]"
                          (mistty--prompt-source prompt)
@@ -1349,7 +1440,7 @@ Return nil if the row isn't reachable on the terminal."
              (setf (mistty--prompt-end prompt) (mistty--term-scrolline)))))
 
         (?D ;; end of command (possible anytime after ?A)
-         (when-let ((prompt (mistty--prompt)))
+         (when-let* ((prompt (mistty--prompt)))
            (when (and (eq 'osc133 (mistty--prompt-source prompt))
                       (null (mistty--prompt-end prompt)))
              (mistty-log "Aborted %s prompt #%s [%s-]"
@@ -1358,6 +1449,15 @@ Return nil if the row isn't reachable on the terminal."
                          (mistty--prompt-start prompt))
              (setf (mistty--prompt-end prompt) (mistty--term-scrolline)))))))))
 
+(defun mistty-call-term-mode-hook ()
+  "Call the functions registered to `term-mode-hook'.
+
+Remove this hook from `mistty-term-mode-hook' to allow the terminal
+modes started by MisTTY to have a completely separate setup from normal
+terminal modes. See the documentation of `mistty-term-mode-hook' for
+details."
+  (run-hooks 'mistty-shadowed-term-mode-hook))
+
 (provide 'mistty-term)
 
 ;;; mistty-term.el ends here
